format PE GUI 4.0 at 0x00400000

entry EntryPoint

include 'win32axp.inc'

; must be at 0x00401000
section '.jtext' readable executable
jpegdec_code file 'inc/jpegdec/jcode.dmp'
JPEG_Decode = jpegdec_code + 0x00e0
JPEG_Free   = jpegdec_code + 0x0150

; must be at 0x00404000
section '.jdata' readable writeable
jpegdec_data file 'inc/jpegdec/jdata.dmp'
jpegdec_VirtualAlloc = jpegdec_data + 0x0000
jpegdec_VirtualFree  = jpegdec_data + 0x0004

.code

proc jpegdec_Init
     push dword [VirtualAlloc]
     pop dword [jpegdec_VirtualAlloc]
     push dword [VirtualFree]
     pop dword [jpegdec_VirtualFree]
     ret
endp

include 'inc/jpegdec/equates.inc'
include 'inc/jpegdec/jpegdec.inc'
include 'inc/olepict.inc'
include 'inc/ddraw.inc'

; common constants
NUMBER_OF_FRAMES equ 150
FRAME_RATE equ 30
START_FRAME equ 0
VSYNC_ENABLED equ FALSE ; big CPU usage if enabled and FRAME_RATE greater than 15

proc EntryPoint
     invoke GetModuleHandle,NULL
     mov [hInstance],eax
     invoke InitCommonControls
     invoke CoInitializeEx,NULL,COINIT_MULTITHREADED+COINIT_SPEED_OVER_MEMORY
     invoke DialogBoxParam,[hInstance],ID_DIALOG_MAIN,HWND_DESKTOP,MainDlgProc,0
     invoke CoUninitialize
     invoke ExitProcess,0
endp

;===========================================================;
proc MainDlgProc uses ebx esi edi,hWndDlg,uMsg,wParam,lParam
;===========================================================;
     locals
       image dd ?
       jpegError dd ?
     endl

     cmp [uMsg],WM_INITDIALOG
     je .wm_initdlg

     cmp [uMsg],WM_COMMAND
     je .wm_command

     cmp [uMsg],WM_ERASEBKGND
     je .wm_erasebkgnd

     cmp [uMsg],WM_LBUTTONDOWN
     je .wm_lbtndown

     cmp [uMsg],WM_LBUTTONDBLCLK
     je .wm_lbuttondblclk

     cmp [uMsg],WM_SETCURSOR
     je .wm_setcursor

     cmp [uMsg],WM_CLOSE
     je .wm_close

     xor eax,eax
     jmp .exit
;------------------------------------------;
.wm_initdlg:
;------------------------------------------;
     invoke GetCurrentThread
     invoke SetThreadPriority,eax,THREAD_PRIORITY_ABOVE_NORMAL

     push [hWndDlg]
     pop [hWnd]

     invoke LoadImage,[hInstance],ID_MAIN_ICON,IMAGE_ICON,0,0,0
     .if eax
	 mov [hIcon],eax
	 invoke SendMessage,[hWndDlg],WM_SETICON,ICON_BIG+ICON_SMALL,eax
     .endif

     invoke LoadImage,[hInstance],ID_NORMAL_CURSOR,IMAGE_CURSOR,0,0,0
     mov [hNormalCur],eax
     invoke LoadImage,[hInstance],ID_MOVE_CURSOR,IMAGE_CURSOR,0,0,0
     mov [hMoveCur],eax

if VSYNC_ENABLED=TRUE
     ; create DirectDraw object for VSYNC.
     invoke DirectDrawCreate,NULL,DDraw,NULL
     .if eax <> DD_OK
	 xor eax,eax
	 mov [DDraw],eax
     .endif
end if

     call jpegdec_Init

     and [bFastMethod],0

     ; note: Jpeg_Load uses SSE and not supports a progressive encoded JPEG files
     ; see return value in jpegError if ->EAX==0
     stdcall Jpeg_Load,[hInstance],ID_FRAME_BASE,RT_RCDATA,addr jpegError
     .if eax
	 mov [image],eax

	 assume eax,JPEGIMAGE

	 push [.eax.width]
	 pop [clientWidth]
	 push [.eax.height]
	 pop [clientHeight]

	 stdcall ResizeWindowAndCenterToScreen,[hWndDlg],[clientWidth],[clientHeight],FALSE

	 stdcall Jpeg_CreateBitmap,[hWndDlg],[image]
	 .if eax
	     mov [pPrevPic],eax

	     invoke SendDlgItemMessage,[hWndDlg],ID_PAINT_BOX,STM_SETIMAGE,IMAGE_BITMAP,[pPrevPic]

	     invoke CreateThread,NULL,0,PaintThreadProc,PaintTimerProcFast,0,threadId
	     .if eax
		 mov [threadHandle],eax
		 ;invoke SetThreadPriority,eax,THREAD_PRIORITY_HIGHEST

		 stdcall uFMOD_PlaySong,pSound,pSound_length,XM_MEMORY
	     .endif

	     or [bFastMethod],-1
	 .endif

	 stdcall Jpeg_Free,[image]
     .else
	 ;invoke DebugBreak
	 ;mov eax,[jpegError]

	 stdcall IPicture_Load,[hInstance],ID_FRAME_BASE,RT_RCDATA
	 .if eax
	     mov [pPrevPic],eax

	     stdcall IPicture_GetSize,[pPrevPic],clientWidth,clientHeight
	     .if eax
		 stdcall ResizeWindowAndCenterToScreen,[hWndDlg],[clientWidth],[clientHeight],FALSE

		 stdcall IPicture_GetBitmapHandle,[pPrevPic]
		 .if eax
		     invoke SendDlgItemMessage,[hWndDlg],ID_PAINT_BOX,STM_SETIMAGE,IMAGE_BITMAP,eax

		     invoke CreateThread,NULL,0,PaintThreadProc,PaintTimerProcCompatible,0,threadId
		     .if eax
			 mov [threadHandle],eax
			 ;invoke SetThreadPriority,eax,THREAD_PRIORITY_HIGHEST

			 stdcall uFMOD_PlaySong,pSound,pSound_length,XM_MEMORY
		     .endif

		 .endif
	     .endif
	 .endif
     .endif

     xor eax,eax
     jmp .exit
;------------------------------------------;
.wm_command:
;------------------------------------------;
     cmp [wParam],BN_CLICKED shl 16 or IDCANCEL
     je .wm_close

     xor eax,eax
     jmp .exit
;------------------------------------------;
.wm_erasebkgnd:
;------------------------------------------;
     xor eax,eax
     inc eax
     jmp .exit
;------------------------------------------;
.wm_lbtndown:
;------------------------------------------;
     invoke SetCursor,[hMoveCur]
     invoke SendMessage,[hWndDlg],WM_NCLBUTTONDOWN,2,0

     xor eax,eax
     jmp .exit
;------------------------------------------;
.wm_lbuttondblclk:
;------------------------------------------;
     not [bDrawLock]

     xor eax,eax
     jmp .exit
;------------------------------------------;
.wm_setcursor:
;------------------------------------------;
     invoke SetCursor,[hNormalCur]

     xor eax,eax
     inc eax
     jmp .exit
;------------------------------------------;
.wm_close:
;------------------------------------------;
     stdcall uFMOD_PlaySong,NULL,0,0

     .if [hShutdownEvent]
	 invoke SetEvent,[hShutdownEvent]
	 invoke Sleep,100
	 invoke TerminateThread,[threadHandle],0
     .endif

     .if [hNormalCur]
	 invoke DestroyCursor,[hNormalCur]
     .endif
     .if [hMoveCur]
	 invoke DestroyCursor,[hMoveCur]
     .endif

     .if [hIcon]
	 invoke DestroyIcon,[hIcon]
     .endif

     .if [pPrevPic]
	 .if [bFastMethod]
	     invoke DeleteObject,[pPrevPic]
	 .else
	     stdcall IPicture_Free,[pPrevPic]
	 .endif
     .endif

if VSYNC_ENABLED=TRUE
     .if [DDraw]
	 cominvk DDraw,Release
     .endif
end if

     invoke EndDialog,[hWndDlg],NULL
     xor eax,eax
;------------------------------------------;
.exit:
;------------------------------------------;
     ret
endp

proc ResizeWindowAndCenterToScreen hWndDlg,NewClientWidth,NewClientHeight,WndHasMenu

     local WndRect:RECT

     xor eax,eax
     mov [WndRect.left],eax
     mov [WndRect.top],eax
     push [NewClientWidth]
     pop [WndRect.right]
     push [NewClientHeight]
     pop [WndRect.bottom]

     invoke GetWindowLong,[hWndDlg],GWL_STYLE
     push eax
     invoke GetWindowLong,[hWndDlg],GWL_EXSTYLE
     pop edx
     invoke AdjustWindowRectEx,addr WndRect,edx,[WndHasMenu],eax

     mov eax,[WndRect.right]
     sub eax,[WndRect.left]
     mov [WndRect.right],eax

     mov eax,[WndRect.bottom]
     sub eax,[WndRect.top]
     mov [WndRect.bottom],eax

     invoke GetSystemMetrics,SM_CXSCREEN
     sub eax,[WndRect.right]
     shr eax,1
     mov [WndRect.left],eax

     invoke GetSystemMetrics,SM_CYSCREEN
     sub eax,[WndRect.bottom]
     shr eax,1
     mov [WndRect.top],eax

     invoke MoveWindow,[hWndDlg],[WndRect.left],[WndRect.top],[WndRect.right],[WndRect.bottom],FALSE
     ret
endp

proc PaintThreadProc lpParameter

     locals
	hTimer dd ?
	DueTime LARGE_INTEGER
	Result dd ?
     endl

     mov [Result],0

     invoke CreateWaitableTimer,NULL,FALSE,NULL
     .if eax
	 mov [hTimer],eax
	 invoke CreateEvent,NULL,TRUE,FALSE,NULL
	 .if eax
	     mov [hShutdownEvent],eax

	     xor eax,eax
	     mov [DueTime.LowPart],eax
	     mov [DueTime.HighPart],eax
	     invoke SetWaitableTimer,[hTimer],addr DueTime,(1000/FRAME_RATE),[lpParameter],NULL,FALSE
	     .if eax
		 .repeat
		     invoke WaitForSingleObject,[hShutdownEvent],0
		     cmp eax,WAIT_OBJECT_0
		     je @f
		     invoke SleepEx,DWORD(-1),TRUE
		 .until FALSE
@@:		 invoke CancelWaitableTimer,[hTimer]
	     .endif

	     invoke CloseHandle,[hShutdownEvent]
	 .endif
	 invoke CloseHandle,[hTimer]
     .endif

     invoke ExitThread,[Result]
     ret
endp

proc PaintTimerProcFast lpArgToCompletionRoutine,dwTimerLowValue,dwTimerHighValue

     locals
	image dd ?
	jpegError dd ?
     endl

     cmp [bDrawLock],0
     jne .exit
     or [bDrawLock],-1

if VSYNC_ENABLED=TRUE
     .if [DDraw]
	 cominvk DDraw,WaitForVerticalBlank,DDWAITVB_BLOCKBEGIN,NULL
     .endif
end if

     mov eax,ID_FRAME_BASE
     add eax,[dwCurrentFrame]

     stdcall Jpeg_Load,[hInstance],eax,RT_RCDATA,addr jpegError
     .if eax
	 mov [image],eax

	 stdcall Jpeg_CreateBitmap,[hWnd],eax
	 .if eax
	     push eax
	     .if [pPrevPic]
		 invoke DeleteObject,[pPrevPic]
	     .endif
	     pop [pPrevPic]

	     invoke SendDlgItemMessage,[hWnd],ID_PAINT_BOX,STM_SETIMAGE,IMAGE_BITMAP,[pPrevPic]
	 .endif

	 stdcall Jpeg_Free,[image]
     .endif

     .if [bDrawReverse]
	 dec [dwCurrentFrame]
     .else
	 inc [dwCurrentFrame]
     .endif

     cmp [dwCurrentFrame],NUMBER_OF_FRAMES
     jb .001
     not [bDrawReverse]
     cmp [dwCurrentFrame],0
     jl @f
     dec [dwCurrentFrame]
     jmp .001
@@:  inc [dwCurrentFrame]
.001:

     and [bDrawLock],0
.exit:

     ret
endp

proc PaintTimerProcCompatible lpArgToCompletionRoutine,dwTimerLowValue,dwTimerHighValue

     locals
	vsync dd ?
     endl

     cmp [bDrawLock],0
     jne .exit
     or [bDrawLock],-1

if VSYNC_ENABLED=TRUE
     .if [DDraw]
	 cominvk DDraw,WaitForVerticalBlank,DDWAITVB_BLOCKBEGIN,NULL
     .endif
end if

     mov eax,ID_FRAME_BASE
     add eax,[dwCurrentFrame]

     stdcall IPicture_Load,[hInstance],eax,RT_RCDATA
     .if eax
	 push eax
	 .if [pPrevPic]
	     stdcall IPicture_Free,[pPrevPic]
	 .endif
	 pop [pPrevPic]

	 stdcall IPicture_GetBitmapHandle,[pPrevPic]
	 .if eax
	     invoke SendDlgItemMessage,[hWnd],ID_PAINT_BOX,STM_SETIMAGE,IMAGE_BITMAP,eax
	 .endif
     .endif

     .if [bDrawReverse]
	 dec [dwCurrentFrame]
     .else
	 inc [dwCurrentFrame]
     .endif

     cmp [dwCurrentFrame],NUMBER_OF_FRAMES
     jb .001
     not [bDrawReverse]
     cmp [dwCurrentFrame],0
     jl @f
     dec [dwCurrentFrame]
     jmp .001
@@:  inc [dwCurrentFrame]
.001:

     and [bDrawLock],0
.exit:

     ret
endp

; ##################################################################################

.data

     hInstance dd ?
     hWnd dd ?

     clientWidth dd ?
     clientHeight dd ?

     hIcon dd 0

     hNormalCur dd 0
     hMoveCur dd 0

     bFastMethod db ?

     threadId dd ?
     threadHandle dd 0
     threadExitCode dd ?
     hShutdownEvent dd 0

     pPrevPic dd 0
     dwCurrentFrame dd START_FRAME

if VSYNC_ENABLED=TRUE
     DDraw IDirectDraw
end if

     bDrawReverse db 0
     bDrawLock db 0

     pSound file 'media/loop.xm'
     pSound_length = $-pSound

; ##################################################################################

section '.idata' data import readable writeable

     library kernel32,'KERNEL32.DLL',\
	     user32,'USER32.DLL',\
	     gdi32,'GDI32.DLL',\
	     ole32,'OLE32.DLL',\
	     oleaut32,'OLEAUT32.DLL',\
	     comctl32,'COMCTL32.DLL',\
	     ddraw,'DDRAW.DLL',\
	     winmm,'WINMM.DLL'

     import_kernel32
     import_user32
     import_gdi32
     import_ole32
     import_comctl32
     import_winmm

     import ddraw,\
	    DirectDrawCreate,'DirectDrawCreate'

     import oleaut32,\
	    OleLoadPicture,'OleLoadPicture'

     all_api

; this section must be after imports
section '.ufmod' code readable executable

; uFMOD constants:
XM_RESOURCE	  = 0
XM_MEMORY	  = 1
XM_FILE 	  = 2
XM_NOLOOP	  = 8
XM_SUSPENDED	  = 16
uFMOD_MIN_VOL	  = 0
uFMOD_MAX_VOL	  = 25
uFMOD_DEFAULT_VOL = 25

macro PUBLIC symbol {} ; hide all publics
macro thnk nm { nm equ [nm] }

; uFMOD setup:
UF_OUTPUT equ WINMM   ; Set output driver to WINMM       (WINMM, OPENAL, DIRECTSOUND)
UF_FREQ   equ 44100   ; Set sampling rate to 44KHz       (22050, 44100, 48000)
UF_RAMP   equ STRONG  ; Select STRONG interpolation      (NONE, WEAK, STRONG)
UF_UFS	  equ UNICODE ; Select UNICODE encoding          (ANSI, UNICODE)
UD_MODE   equ NORMAL  ; Select NORMAL mode               (NORMAL, UNSAFE, BENCHMARK)
VB6	  equ 0       ; Disable VisualBasic 6 extensions
PBASIC	  equ 0       ; Disable PureBasic extensions
EBASIC	  equ 0       ; Disable Emergence BASIC extensions
BLITZMAX  equ 0       ; Disable BlitzMax extensions
NOLINKER  equ 1       ; Select "no linker" mode
UF_FMT	  equ 0       ; Disable DLL entry point

include 'inc/eff.inc'
include 'inc/ufmod/fasm.asm'

; ##################################################################################

section '.rsrc' resource data readable

     ID_DIALOG_MAIN   = 100
     ID_PAINT_BOX     = 200
     ID_FRAME_BASE    = 400
     ID_MAIN_ICON     = 500
     ID_NORMAL_CURSOR = 800
     ID_MOVE_CURSOR   = 801

     directory RT_DIALOG,_dialogs,\
	       RT_GROUP_ICON,_groupicons,\
	       RT_ICON,_icons,\
	       RT_GROUP_CURSOR,_groupcursors,\
	       RT_CURSOR,_cursors,\
	       RT_RCDATA,_rcdata

     resource _dialogs,\
	      ID_DIALOG_MAIN,LANG_ENGLISH+SUBLANG_ENGLISH_US,_main_dlg

     resource _groupicons,\
	      ID_MAIN_ICON,LANG_NEUTRAL,main_icon

     resource _icons,\
	      ID_MAIN_ICON,LANG_NEUTRAL,main_icon_data

     resource _groupcursors,\
	      ID_NORMAL_CURSOR,LANG_NEUTRAL,normal_cursor,\
	      ID_MOVE_CURSOR,LANG_NEUTRAL,move_cursor

     resource _cursors,\
	      ID_NORMAL_CURSOR,LANG_NEUTRAL,normal_cursor_data,\
	      ID_MOVE_CURSOR,LANG_NEUTRAL,move_cursor_data

     ; generate this file with ResAuto tool!
     include 'inc/frames.inc'

     dialog _main_dlg,'JPEG animation example',0,0,160,120,WS_POPUP+WS_BORDER+WS_SYSMENU+DS_CENTER
	    dialogitem 'STATIC','',ID_PAINT_BOX,0,0,0,0,WS_VISIBLE+SS_BITMAP
     enddialog

     icon main_icon,main_icon_data,'media/icon.ico'

     cursor normal_cursor,normal_cursor_data,'media/normal.cur'
     cursor move_cursor,move_cursor_data,'media/move.cur'

     rept NUMBER_OF_FRAMES num:1 {
	  ID_IMAGE#num = ID_FRAME_BASE+num-1
	  resdata _img#num
		  file 'media/frames/'#`num#'.jpg'
	  endres
     }

; ##################################################################################
